from spinecore.common.context.base_context import BaseContext
from spinecore.common.helpers.contract.Workflow import Workflow


class IExecutor:
    async def execute(self, ctx: BaseContext, workflow: Workflow) -> dict:
        raise NotImplementedError()



import uuid
from typing import List

from spinecore.common.helpers.contract.IContext import IContext
from spinecore.common.helpers.logging import SpineLogger
from spinecore.common.helpers.contract import Logger
from spinecore.common.helpers.contract.metaclasses.singleton import Singleton


class BaseContext(IContext, metaclass=Singleton):
    """Base spine context"""

    def __init__(self, logger: Logger = SpineLogger, log_handlers: List = []):
        self.run_id = str(uuid.uuid1()).lower()
        self.__spine_logger = logger(self.run_id, log_handlers)
        self.logger = self.__spine_logger.logger

    def get_spine_logger(self) -> SpineLogger:
        return self.__spine_logger

    @staticmethod
    def get_context(log_handlers: List = []):
        ctx = BaseContext(log_handlers=log_handlers)
        return ctx



import logging
import sys
import time
from functools import wraps

from spinecore.common.helpers.contract.Logger import Logger


class SpineLogger(Logger):
    def __init__(self, run_id: str, handlers=None, name: str = "SpineLogger"):
        super().__init__(run_id)
        handlers = handlers if handlers else []
        self.handlers = handlers
        self.__setup_logger()
        self.logger = logging.getLogger(name)

    def __setup_logger(self):
        logging.basicConfig(level=logging.INFO, handlers=[logging.StreamHandler(sys.stdout)] + self.handlers,
                            format=f'{self.run_id} | %(asctime)s %(levelname)s | %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S')

    def time_function(self, func, msg: str):
        @wraps(func)
        async def step_wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = await func(*args, **kwargs)
            end_time = time.perf_counter()
            self.logger.info(f"{msg} Took {end_time - start_time:.4f} seconds")
            return result

        return step_wrapper



from typing import Union, Tuple

from spinecore.common.context.spark_context import SpineSparkContext
from spinecore.common.helpers.dataframes.spark_dataframe import SparkDataFrame
from spinecore.common.helpers.contract.IStep import IStep


class SparkStep(IStep):
    """IStep implementation for spark steps"""

    async def process(self, ctx: SpineSparkContext, df: Union[SparkDataFrame, Tuple[SparkDataFrame]]) \
            -> Union[SparkDataFrame, Tuple[SparkDataFrame]]:
        raise NotImplementedError



from typing import List, Iterable

import networkx as nx

from spinecore.common.helpers.contract.IStep import IStep
from spinecore.common.helpers.contract.Workflow import Workflow
from spinecore.common.helpers.graphs import topological_pos, default_dag_style


class DagWorkflow(Workflow):
    """DAG (Directed Acyclic Graph) Implementation of Workflow"""

    def __init__(self, workflow_name: str, graph=nx.DiGraph()):
        super().__init__(workflow_name, graph=graph)

    def add_after(self, next_steps: List[IStep], prev_step: List[IStep]) -> None:
        self.graph.add_nodes_from(next_steps)
        for next_step in next_steps:
            self.graph.add_edges_from([(prev_step, next_step) for prev_step in prev_step])

    def add_last(self, *next_steps: IStep) -> None:
        leaf_nodes = self.__get_leaf_nodes()
        self.graph.add_nodes_from(next_steps)
        if leaf_nodes:
            for node in next_steps:
                self.graph.add_edges_from([(leaf_node, node)
                                           for leaf_node in leaf_nodes])

    def __get_leaf_nodes(self) -> Iterable[List]:
        """Returns an iterator over leaf in current graph"""
        return iter([node for node in self.graph.nodes()
                     if self.graph.out_degree(node) == 0])

    def flat(self) -> List:
        """Returns a generator of nodes in topologically sorted order"""
        return nx.algorithms.topological_sort(self.graph)

    def get_all_sub_workflows(self) -> Iterable:
        """Returns all the sub workflows"""
        return iter([DagWorkflow(f"{self.name}", self.graph.subgraph(c).copy()) for c in
                     nx.weakly_connected_components(self.graph)])

    def get_before(self, step: IStep = None):
        """Returns an iterator over predecessor nodes of n"""
        if step is None:
            return self.__get_leaf_nodes()
        return self.graph.predecessors(step)

    def show_workflow(self, style: dict = None) -> None:
        try:
            from matplotlib import pyplot as plt
            from matplotlib import rcParams
            rcParams.update({'figure.autolayout': True})
            style = style if style else default_dag_style
            nx.draw_networkx(self.graph, pos=topological_pos(self.graph), **style)
            plt.show()
        except ImportError:
            raise ImportError("matplotlib must be installed in order to show graph")



from typing import Union, Tuple

from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.dataframes.polars_dataframe import PolarsDataFrame
from spinecore.common.helpers.contract.IStep import IStep


class PolarsStep(IStep):
    """IStep implementation for polars steps"""

    async def process(self, ctx: SpinePolarsContext, df: Union[PolarsDataFrame, Tuple[PolarsDataFrame]]) \
            -> Union[PolarsDataFrame, Tuple[PolarsDataFrame]]:
        raise NotImplementedError



import pyspark
from pyspark.sql import SparkSession

from spinecore.common.context.base_context import BaseContext


class SpineSparkContext(BaseContext):
    def __init__(self, spark_session: SparkSession, spark_config: dict):
        super().__init__()
        self.spark = self.__setup_spark_session(spark_session, spark_config)

    @staticmethod
    def __setup_spark_session(spark_session: SparkSession, spark_config: dict):
        if spark_session:
            return spark_session
        return SparkSession.builder.config(conf=pyspark.SparkConf().setAll(spark_config.items())) \
            .enableHiveSupport().getOrCreate()

    @staticmethod
    def get_context(spark_config=None, spark_session: SparkSession = None):
        """Get or create SpineSparkContext"""
        spark_config = {} if not spark_config else spark_config
        ctx = SpineSparkContext(spark_session, spark_config)
        ctx.logger.info("Starting SpineSparkContext")
        return ctx

    @staticmethod
    def into_polars(lazy=True, config: dict = None):
        """Convert SpineSparkContext to SpinePolarsContext"""
        try:
            from spinecore.common.context.polars_context import SpinePolarsContext
            return SpinePolarsContext(lazy, config)
        except ImportError:
            raise ImportError("Polars must be installed in order to use PolarsSparkContext")



import asyncio

import polars as pl
import pytest

from spinecore.common.context.polars_context import SpinePolarsContext
from spinecore.common.helpers.contract.IStep import IStep
from spinecore.executors import execute
from spinecore.workflows import DagWorkflow


class BlockingStep(IStep):
    async def process(self, ctx: SpinePolarsContext, df=None):
        df = pl.DataFrame({"foo": [1, 2, 3], "bar": [None, "ham", "spam"]})
        await asyncio.sleep(1)  # Blocking
        return df


class NothingStep(IStep):
    async def process(self, ctx: SpinePolarsContext, df):
        return df


@pytest.mark.asyncio
async def test_blocking_step():
    """Tests IO Blocking step execution"""
    # Prepare
    ctx = SpinePolarsContext.get_context(lazy=True, config={"set_tbl_rows": 20})
    workflow = DagWorkflow("test_blocking")
    expected_df = pl.DataFrame({"foo": [1, 2, 3], "bar": [None, "ham", "spam"]})
    nothing_step = NothingStep()
    workflow.add_after([nothing_step], [BlockingStep()])

    # Action
    result = await execute(ctx, workflow)
    df = result[0][nothing_step]

    # Assert
    assert df.frame_equal(expected_df)



from spinecore.common.context.base_context import BaseContext
from spinecore.common.helpers.contract.IExecutor import IExecutor
from spinecore.common.helpers.contract.Workflow import Workflow
from spinecore.executors.workflow_executors.async_workflow_executor import AsyncWorkflowExecutor


async def execute(ctx: BaseContext, workflow: Workflow, executor: IExecutor = AsyncWorkflowExecutor()) -> dict:
    """Executes workflow using ctx and executor"""
    return await ctx.get_spine_logger().time_function(executor.execute, f"{workflow.name} execution")(ctx, workflow)



from typing import Union, Tuple

from spinecore.common.context.pandas_context import SpinePandasContext
from spinecore.common.helpers.contract.IStep import IStep
from spinecore.common.helpers.dataframes.pandas_dataframe import PandasDataFrame


class PandasStep(IStep):
    """IStep implementation for pandas steps"""

    async def process(self, ctx: SpinePandasContext, df: Union[PandasDataFrame, Tuple[PandasDataFrame]]) \
            -> Union[PandasDataFrame, Tuple[PandasDataFrame]]:
        raise NotImplementedError



from spinecore.common.context.base_context import BaseContext


class SpinePolarsContext(BaseContext):
    def __init__(self, lazy: bool, config: dict):
        super().__init__()
        self.lazy = lazy
        self.__setup_config(config)

    @staticmethod
    def __setup_config(config: dict):
        """Sets polars config from dict"""
        # TODO: Make this less ugly
        if config:
            exec("import polars as pl")
            for key, value in config.items():
                exec(f"pl.Config.{key}({value})")

    @staticmethod
    def get_context(lazy: bool = True, config: dict = None):
        """Get or create SpinePolars context"""
        ctx = SpinePolarsContext(lazy, config)
        ctx.logger.info(f"Starting SpinePolarsContext with lazy_mode={lazy}")
        return ctx

    @staticmethod
    def into_spark(spark_config: dict = {}, spark_session=None):
        """Converts SpinePolarsContext to SpineSparkContext"""
        try:
            spark_config = {} if not spark_config else spark_config
            from spinecore.common.context.spark_context import SpineSparkContext
            return SpineSparkContext(spark_session, spark_config)
        except ImportError:
            raise ImportError("Pyspark must be installed in order to use SpineSparkContext")
